#!/usr/bin/env python
# -*- coding: utf-8 -*-

__version__ = '1.2.1'

# Copyright (C) 2014,2015,2016,2017,2018 Cumulus Networks, Inc. All rights reserved
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#
# https://www.gnu.org/licenses/gpl-2.0-standalone.html
#
# Authors:
#       Roopa Prabhu, roopa@cumulusnetworks.com
#       Julien Fortin, julien@cumulusnetworks.com

__license__ = 'GPL-2'
__status__ = 'Production'
__copyright__ = 'Copyright (C) 2014,2015,2016,2017,2018 Cumulus Networks, Inc.'


__authors__ = [
    'Roopa Prabhu',
    'Julien Fortin'
]
__credits__ = [
    'Roopa Prabhu',
    'Julien Fortin'
]
__emails__ = [
    'roopa@cumulusnetworks.com',
    'julien@cumulusnetworks.com'
]
__maintainer__ = 'Julien Fortin'
__email__ = 'julien@cumulusnetworks.com'
